package solutions.bitbadger.documents.scala

import solutions.bitbadger.documents.{DocumentIndex, Field, FieldMatch, Parameter}

import java.sql.{Connection, ResultSet}
import scala.reflect.ClassTag

package object extensions:

  extension (conn: Connection)

    // ~~~ CUSTOM QUERIES ~~~

    /**
     * Execute a query that returns a list of results
     *
     * @param query      The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc    The mapping function between the document and the domain item
     * @return A list of results for the given query
     * @throws DocumentException If parameters are invalid
     */
    def customList[TDoc](query: String, parameters: Seq[Parameter[?]],
                         mapFunc: (ResultSet, ClassTag[TDoc]) => TDoc)(implicit tag: ClassTag[TDoc]): List[TDoc] =
      Custom.list[TDoc](query, parameters, conn, mapFunc)

    /**
     * Execute a query that returns a list of results
     *
     * @param query   The query to retrieve the results
     * @param mapFunc The mapping function between the document and the domain item
     * @return A list of results for the given query
     * @throws DocumentException If parameters are invalid
     */
    def customList[TDoc](query: String,
                         mapFunc: (ResultSet, ClassTag[TDoc]) => TDoc)(implicit tag: ClassTag[TDoc]): List[TDoc] =
      Custom.list[TDoc](query, conn, mapFunc)

    /**
     * Execute a query that returns one or no results
     *
     * @param query      The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @param mapFunc    The mapping function between the document and the domain item
     * @return An optional document, filled if one matches the query
     * @throws DocumentException If parameters are invalid
     */
    def customSingle[TDoc](query: String, parameters: Seq[Parameter[?]],
                           mapFunc: (ResultSet, ClassTag[TDoc]) => TDoc)(implicit tag: ClassTag[TDoc]): Option[TDoc] =
      Custom.single[TDoc](query, parameters, conn, mapFunc)

    /**
     * Execute a query that returns one or no results
     *
     * @param query   The query to retrieve the results
     * @param mapFunc The mapping function between the document and the domain item
     * @return An optional document, filled if one matches the query
     * @throws DocumentException If parameters are invalid
     */
    def customSingle[TDoc](query: String,
                           mapFunc: (ResultSet, ClassTag[TDoc]) => TDoc)(implicit tag: ClassTag[TDoc]): Option[TDoc] =
      Custom.single[TDoc](query, conn, mapFunc)

    /**
     * Execute a query that returns no results
     *
     * @param query      The query to retrieve the results
     * @param parameters Parameters to use for the query
     * @throws DocumentException If parameters are invalid
     */
    def customNonQuery(query: String, parameters: Seq[Parameter[?]] = List()): Unit =
      Custom.nonQuery(query, parameters, conn)

    /**
     * Execute a query that returns a scalar result
     *
     * @param query      The query to retrieve the result
     * @param parameters Parameters to use for the query
     * @param mapFunc    The mapping function between the document and the domain item
     * @return The scalar value from the query
     * @throws DocumentException If parameters are invalid
     */
    def customScalar[T](query: String, parameters: Seq[Parameter[?]],
                        mapFunc: (ResultSet, ClassTag[T]) => T)(implicit tag: ClassTag[T]): T =
      Custom.scalar[T](query, parameters, conn, mapFunc)

    /**
     * Execute a query that returns a scalar result
     *
     * @param query   The query to retrieve the result
     * @param mapFunc The mapping function between the document and the domain item
     * @return The scalar value from the query
     * @throws DocumentException If parameters are invalid
     */
    def customScalar[T](query: String,
                        mapFunc: (ResultSet, ClassTag[T]) => T)(implicit tag: ClassTag[T]): T =
      Custom.scalar[T](query, conn, mapFunc)

    // ~~~ DEFINITION QUERIES ~~~

    /**
     * Create a document table if necessary
     *
     * @param tableName The table whose existence should be ensured (may include schema)
     * @throws DocumentException If the dialect is not configured
     */
    def ensureTable(tableName: String): Unit =
      Definition.ensureTable(tableName, conn)

    /**
     * Create an index on field(s) within documents in the specified table if necessary
     *
     * @param tableName The table to be indexed (may include schema)
     * @param indexName The name of the index to create
     * @param fields    One or more fields to be indexed<
     * @throws DocumentException If any dependent process does
     */
    def ensureFieldIndex(tableName: String, indexName: String, fields: Seq[String]): Unit =
      Definition.ensureFieldIndex(tableName, indexName, fields, conn)

    /**
     * Create a document index on a table (PostgreSQL only)
     *
     * @param tableName The table to be indexed (may include schema)
     * @param indexType The type of index to ensure
     * @throws DocumentException If called on a SQLite connection
     */
    def ensureDocumentIndex(tableName: String, indexType: DocumentIndex): Unit =
      Definition.ensureDocumentIndex (tableName, indexType, conn)

    // ~~~ DOCUMENT MANIPULATION QUERIES ~~~

    /**
     * Insert a new document
     *
     * @param tableName The table into which the document should be inserted (may include schema)
     * @param document  The document to be inserted
     * @throws DocumentException If IDs are misconfigured, or if the database command fails
     */
    def insert[TDoc](tableName: String, document: TDoc): Unit =
      Document.insert(tableName, document, conn)

    /**
     * Save a document, inserting it if it does not exist and updating it if it does (AKA "upsert")
     *
     * @param tableName The table in which the document should be saved (may include schema)
     * @param document  The document to be saved
     * @throws DocumentException If the database command fails
     */
    def save[TDoc](tableName: String, document: TDoc): Unit =
      Document.save(tableName, document, conn)

    /**
     * Update (replace) a document by its ID
     *
     * @param tableName The table in which the document should be replaced (may include schema)
     * @param docId     The ID of the document to be replaced
     * @param document  The document to be replaced
     * @throws DocumentException If no dialect has been configured, or if the database command fails
     */
    def update[TKey, TDoc](tableName: String, docId: TKey, document: TDoc): Unit =
      Document.update(tableName, docId, document, conn)

    // ~~~ DOCUMENT COUNT QUERIES ~~~

    /**
     * Count all documents in the table
     *
     * @param tableName The name of the table in which documents should be counted
     * @return A count of the documents in the table
     * @throws DocumentException If any dependent process does
     */
    def countAll(tableName: String): Long =
      Count.all(tableName, conn)

    /**
     * Count documents using a field comparison
     *
     * @param tableName  The name of the table in which documents should be counted
     * @param fields     The fields which should be compared
     * @param howMatched How the fields should be matched (optional, default `ALL`)
     * @return A count of the matching documents in the table
     * @throws DocumentException If the dialect has not been configured
     */
    def countByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Long =
      Count.byFields(tableName, fields, howMatched, conn)

    /**
     * Count documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which documents should be counted
     * @param criteria  The object for which JSON containment should be checked
     * @return A count of the matching documents in the table
     * @throws DocumentException If called on a SQLite connection
     */
    def countByContains[TContains](tableName: String, criteria: TContains): Long =
      Count.byContains(tableName, criteria, conn)

    /**
     * Count documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which documents should be counted
     * @param path      The JSON path comparison to match
     * @return A count of the matching documents in the table
     * @throws DocumentException If called on a SQLite connection
     */
    def countByJsonPath(tableName: String, path: String): Long =
      Count.byJsonPath(tableName, path, conn)

    // ~~~ DOCUMENT EXISTENCE QUERIES ~~~

    /**
     * Determine a document's existence by its ID
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param docId     The ID of the document to be checked
     * @return True if the document exists, false if not
     * @throws DocumentException If no dialect has been configured
     */
    def existsById[TKey](tableName: String, docId: TKey): Boolean =
      Exists.byId(tableName, docId, conn)

    /**
     * Determine document existence using a field comparison
     *
     * @param tableName  The name of the table in which document existence should be checked
     * @param fields     The fields which should be compared
     * @param howMatched How the fields should be matched
     * @return True if any matching documents exist, false if not
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    def existsByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Boolean =
      Exists.byFields(tableName, fields, howMatched, conn)

    /**
     * Determine document existence using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param criteria  The object for which JSON containment should be checked
     * @return True if any matching documents exist, false if not
     * @throws DocumentException If called on a SQLite connection
     */
    def existsByContains[TContains](tableName: String, criteria: TContains): Boolean =
      Exists.byContains(tableName, criteria, conn)

    /**
     * Determine document existence using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param path      The JSON path comparison to match
     * @return True if any matching documents exist, false if not
     * @throws DocumentException If called on a SQLite connection
     */
    def existsByJsonPath(tableName: String, path: String): Boolean =
      Exists.byJsonPath(tableName, path, conn)

    // ~~~ DOCUMENT RETRIEVAL QUERIES ~~~

    /**
     * Retrieve all documents in the given table, ordering results by the optional given fields
     *
     * @param tableName The table from which documents should be retrieved
     * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents from the given table
     * @throws DocumentException If query execution fails
     */
    def findAll[TDoc](tableName: String, orderBy: Seq[Field[?]] = List())(implicit tag: ClassTag[TDoc]): List[TDoc] =
      Find.all[TDoc](tableName, orderBy, conn)

    /**
     * Retrieve a document by its ID
     *
     * @param tableName The table from which the document should be retrieved
     * @param docId     The ID of the document to retrieve
     * @return The document if it is found, `None` otherwise
     * @throws DocumentException If no dialect has been configured
     */
    def findById[TKey, TDoc](tableName: String, docId: TKey)(implicit tag: ClassTag[TDoc]): Option[TDoc] =
      Find.byId[TKey, TDoc](tableName, docId, conn)

    /**
     * Retrieve documents using a field comparison, ordering results by the optional given fields
     *
     * @param tableName  The table from which the document should be retrieved
     * @param fields     The fields which should be compared
     * @param howMatched How the fields should be matched
     * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the field comparison
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    def findByFields[TDoc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                           orderBy: Seq[Field[?]] = List())(implicit tag: ClassTag[TDoc]): List[TDoc] =
      Find.byFields[TDoc](tableName, fields, howMatched, orderBy, conn)

    /**
     * Retrieve documents using a JSON containment query, ordering results by the optional given fields (PostgreSQL
     * only)
     *
     * @param tableName The name of the table in which document existence should be checked
     * @param criteria  The object for which JSON containment should be checked
     * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the JSON containment query
     * @throws DocumentException If called on a SQLite connection
     */
    def findByContains[TDoc, TContains](tableName: String, criteria: TContains, orderBy: Seq[Field[?]] = List())
                                       (implicit tag: ClassTag[TDoc]): List[TDoc] =
      Find.byContains[TDoc, TContains](tableName, criteria, orderBy, conn)

    /**
     * Retrieve documents using a JSON Path match query, ordering results by the optional given fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path      The JSON path comparison to match
     * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return A list of documents matching the JSON Path match query
     * @throws DocumentException If called on a SQLite connection
     */
    def findByJsonPath[TDoc](tableName: String, path: String, orderBy: Seq[Field[?]] = List())
                            (implicit tag: ClassTag[TDoc]): List[TDoc] =
      Find.byJsonPath[TDoc](tableName, path, orderBy, conn)

    /**
     * Retrieve the first document using a field comparison and optional ordering fields
     *
     * @param tableName  The table from which documents should be retrieved
     * @param fields     The fields which should be compared
     * @param howMatched How the fields should be matched (optional, defaults to `FieldMatch.ALL`)
     * @param orderBy    Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first document matching the field comparison, or `None` if no matches are found
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    def findFirstByFields[TDoc](tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None,
                                orderBy: Seq[Field[?]] = List())(implicit tag: ClassTag[TDoc]): Option[TDoc] =
      Find.firstByFields[TDoc](tableName, fields, howMatched, orderBy, conn)

    /**
     * Retrieve the first document using a JSON containment query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param criteria  The object for which JSON containment should be checked
     * @param orderBy   Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first document matching the JSON containment query, or `None` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    def findFirstByContains[TDoc, TContains](tableName: String, criteria: TContains, orderBy: Seq[Field[?]] = List())
                                            (implicit tag: ClassTag[TDoc]): Option[TDoc] =
      Find.firstByContains[TDoc, TContains](tableName, criteria, orderBy, conn)

    /**
     * Retrieve the first document using a JSON Path match query and optional ordering fields (PostgreSQL only)
     *
     * @param tableName The table from which documents should be retrieved
     * @param path The JSON path comparison to match
     * @param orderBy Fields by which the query should be ordered (optional, defaults to no ordering)
     * @return The first document matching the JSON Path match query, or `None` if no matches are found
     * @throws DocumentException If called on a SQLite connection
     */
    def findFirstByJsonPath[TDoc](tableName: String, path: String, orderBy: Seq[Field[?]] = List())
                                 (implicit tag: ClassTag[TDoc]): Option[TDoc] =
      Find.firstByJsonPath[TDoc](tableName, path, orderBy, conn)

    // ~~~ DOCUMENT PATCH (PARTIAL UPDATE) QUERIES ~~~

    /**
     * Patch a document by its ID
     *
     * @param tableName The name of the table in which a document should be patched
     * @param docId     The ID of the document to be patched
     * @param patch     The object whose properties should be replaced in the document
     * @throws DocumentException If no dialect has been configured
     */
    def patchById[TKey, TPatch](tableName: String, docId: TKey, patch: TPatch): Unit =
      Patch.byId(tableName, docId, patch, conn)

    /**
     * Patch documents using a field comparison
     *
     * @param tableName  The name of the table in which documents should be patched
     * @param fields     The fields which should be compared
     * @param patch      The object whose properties should be replaced in the document
     * @param howMatched How the fields should be matched
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    def patchByFields[TPatch](tableName: String, fields: Seq[Field[?]], patch: TPatch,
                              howMatched: Option[FieldMatch] = None): Unit =
      Patch.byFields(tableName, fields, patch, howMatched, conn)

    /**
    * Patch documents using a JSON containment query (PostgreSQL only)
    *
    * @param tableName The name of the table in which documents should be patched
    * @param criteria  The object against which JSON containment should be checked
    * @param patch     The object whose properties should be replaced in the document
    * @throws DocumentException If called on a SQLite connection
    */
    def patchByContains[TContains, TPatch](tableName: String, criteria: TContains, patch: TPatch): Unit =
      Patch.byContains(tableName, criteria, patch, conn)

    /**
     * Patch documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table in which documents should be patched
     * @param path The JSON path comparison to match
     * @param patch The object whose properties should be replaced in the document
     * @throws DocumentException If called on a SQLite connection
     */
    def patchByJsonPath[TPatch](tableName: String, path: String, patch: TPatch): Unit =
      Patch.byJsonPath(tableName, path, patch, conn)

    // ~~~ DOCUMENT FIELD REMOVAL QUERIES ~~~

    /**
     * Remove fields from a document by its ID
     *
     * @param tableName The name of the table in which the document's fields should be removed
     * @param docId     The ID of the document to have fields removed
     * @param toRemove  The names of the fields to be removed
     * @throws DocumentException If no dialect has been configured
     */
    def removeFieldsById[TKey](tableName: String, docId: TKey, toRemove: Seq[String]): Unit =
      RemoveFields.byId(tableName, docId, toRemove, conn)

    /**
     * Remove fields from documents using a field comparison
     *
     * @param tableName  The name of the table in which document fields should be removed
     * @param fields     The fields which should be compared
     * @param toRemove   The names of the fields to be removed
     * @param howMatched How the fields should be matched
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    def removeFieldsByFields(tableName: String, fields: Seq[Field[?]], toRemove: Seq[String],
                             howMatched: Option[FieldMatch] = None): Unit =
      RemoveFields.byFields(tableName, fields, toRemove, howMatched, conn)

    /**
     * Remove fields from documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table in which document fields should be removed
     * @param criteria  The object against which JSON containment should be checked
     * @param toRemove  The names of the fields to be removed
     * @throws DocumentException If called on a SQLite connection
     */
    def removeFieldsByContains[TContains](tableName: String, criteria: TContains, toRemove: Seq[String]): Unit =
      RemoveFields.byContains(tableName, criteria, toRemove, conn)

    /**
    * Remove fields from documents using a JSON Path match query (PostgreSQL only)
    *
    * @param tableName The name of the table in which document fields should be removed
    * @param path      The JSON path comparison to match
    * @param toRemove  The names of the fields to be removed
    * @throws DocumentException If called on a SQLite connection
    */
    def removeFieldsByJsonPath(tableName: String, path: String, toRemove: Seq[String]): Unit =
      RemoveFields.byJsonPath(tableName, path, toRemove, conn)

    // ~~~ DOCUMENT DELETION QUERIES ~~~

    /**
     * Delete a document by its ID
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param docId     The ID of the document to be deleted
     * @throws DocumentException If no dialect has been configured
     */
    def deleteById[TKey](tableName: String, docId: TKey): Unit =
      Delete.byId(tableName, docId, conn)

    /**
     * Delete documents using a field comparison
     *
     * @param tableName  The name of the table from which documents should be deleted
     * @param fields     The fields which should be compared
     * @param howMatched How the fields should be matched
     * @throws DocumentException If no dialect has been configured, or if parameters are invalid
     */
    def deleteByFields(tableName: String, fields: Seq[Field[?]], howMatched: Option[FieldMatch] = None): Unit =
      Delete.byFields(tableName, fields, howMatched, conn)

    /**
     * Delete documents using a JSON containment query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param criteria  The object for which JSON containment should be checked
     * @throws DocumentException If called on a SQLite connection
     */
    def deleteByContains[TContains](tableName: String, criteria: TContains): Unit =
      Delete.byContains(tableName, criteria, conn)

    /**
     * Delete documents using a JSON Path match query (PostgreSQL only)
     *
     * @param tableName The name of the table from which documents should be deleted
     * @param path      The JSON path comparison to match
     * @throws DocumentException If called on a SQLite connection
     */
    def deleteByJsonPath(tableName: String, path: String): Unit =
      Delete.byJsonPath(tableName, path, conn)
